#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

typedef struct {
    const char *fn;
    int xs;
    int ys;
    unsigned char *buf;
} pgm;

typedef struct {
    int xs;
    int ys;
    int *sum;
    int *count;
} blendbuf;

static void
die (char *why)
{
  fprintf (stderr, "%s\n", why);
  exit (1);
}

#define MAX_SIZE 65536

pgm *load_pgm(const char *fn)
{
    FILE *fi = fopen(fn, "rb");
    pgm *result;
    char buf[256];
    int xs, ys;
    int depth;

    if (fi == NULL)
	return NULL;

    fgets (buf, sizeof(buf), fi);
    if (buf[0] != 'P' || buf[1] != '5')
	die ("Need pgmraw image on input");

    xs = ys = 0;
    do
	fgets (buf, sizeof(buf), fi);
    while (buf[0] == '#');
    sscanf (buf, "%d %d", &xs, &ys);
    if (xs <= 0 || ys <= 0 || xs > MAX_SIZE || ys > MAX_SIZE)
	die ("Input image size out of range");

    do
	fgets (buf, sizeof(buf), fi);
    while (buf[0] == '#');
    sscanf (buf, "%d", &depth);
    if (depth != 255)
	die ("Only works with depth=255 images");

    result = (pgm *)malloc(sizeof(pgm));

    result->fn = fn;
    result->xs = xs;
    result->ys = ys;
    result->buf = (unsigned char *)malloc(xs * ys);

    fread(result->buf, 1, xs * ys, fi);
    fprintf(stderr, "loaded file %s %dx%d\n", fn, xs, ys);
    fclose(fi);
    return result;
}

int
align_pgms(const pgm *p1, const pgm *p2, int *px, int *py)
{
    int xo, yo;
    int xa, ya;
    int best = 0x7fffffff;

    xa = (p1->xs < p2->xs ? p1->xs : p2->xs) - 20;
    ya = (p1->ys < p2->ys ? p1->ys : p2->ys) - 20;

    for (yo = -10; yo <= 10; yo++)
	for (xo = -10; xo <= 10; xo++) {
	    int sum = 0;
	    int i, j;

	    for (j = 0; j < ya; j++)
		for (i = 0; i < xa; i++) {
		    int g1 = p1->buf[(j + 10) * p1->xs + i + 10];
		    int g2 = p2->buf[(j + 10 - yo) * p2->xs + i - xo + 10];
		    sum += (g1 - g2) * (g1 - g2);
		}
	    if (sum < best) {
		best = sum;
		*px = xo;
		*py = yo;
	    }
	}
    return best;
}

blendbuf *
new_blendbuf(int xs, int ys)
{
    blendbuf *result = (blendbuf *)malloc(sizeof(blendbuf));
    int i;

    result->xs = xs;
    result->ys = ys;
    result->sum = (int *)malloc(sizeof(int) * xs * ys);
    result->count = (int *)malloc(sizeof(int) * xs * ys);
    for (i = 0; i < xs * ys; i++) {
	result->sum[i] = 0;
	result->count[i] = 0;
    }

    return result;
}

void
add_pgm(blendbuf *bb, pgm *p, int xo, int yo)
{
    int i, j;

    for (j = 0; j < p->ys; j++) {
	if (j + yo >= 0 && j + yo < bb->ys) {
	    for (i = 0; i < p->xs; i++) {
		if (i + xo >= 0 && i + xo < bb->xs) {
		    int ix = (j + yo) * bb->xs + i + xo;
		    bb->sum[ix] += p->buf[j * p->xs + i];
		    bb->count[ix]++;
		}
	    }
	}
    }
}

void
print_blendbuf(blendbuf *bb, FILE *f)
{
    int i, j;
    int xs = bb->xs;
    int ys = bb->ys;
    unsigned char *linebuf = malloc(xs);

    fprintf(f, "P5\n%d %d\n255\n", xs, ys);
    for (j = 0; j < ys; j++) {
	for (i = 0; i < xs; i++) {
	    int ix = j * xs + i;
	    unsigned char g;
	    if (bb->count[ix])
		g = (bb->sum[ix] + (bb->count[ix] >> 1)) / bb->count[ix];
	    else
		g = 255;
	    linebuf[i] = g;
	}
	fwrite(linebuf, 1, bb->xs, f);
    }
    free(linebuf);
}

void
threshold(pgm *p)
{
    int i;

    for (i = 0; i < p->xs * p->ys; i++)
	p->buf[i] = p->buf[i] > 128 ? 1 : 0;
}

int 
classify(pgm **pgmlist, int n_pgm)
{
    int *class = (int *)malloc(sizeof(int) * n_pgm);
    int n_class = 0;
    int i, j;
    int tshift = 4;

    for (i = 0; i < n_pgm; i++)
	class[i] = -1;

    for (i = 0; i < n_pgm; i++) {
	pgm *pi = pgmlist[i];

	if (class[i] == -1) {
	    class[i] = n_class++;
	    for (j = i + 1; j < n_pgm; j++) {
		pgm *pj = pgmlist[j];
		int xo, yo;
		int score;

		if (abs(pi->xs - pj->xs) < 10 &&
		    abs(pi->ys - pj->ys) < 10 &&
		    class[j] == -1) {
		    score = align_pgms(pi, pj, &xo, &yo);
		    if (score < ((pi->xs - 20) * (pi->ys - 20)) >> tshift) {
			class[j] = class[i];
		    }
		}
	    }
	}
	printf("%s: class%d\n", pi->fn, class[i]);
	fflush(stdout);
    }
    free(class);
    return 0;
}

#define MAX_FN 65536

int
main(int argc, char **argv)
{
    int i;
    int n_pgm = 0;
    pgm **pgmlist = (pgm **)malloc(sizeof(pgm *) * MAX_FN);
    int do_class = 0;
    blendbuf *bb;

    for (i = 1; i < argc; i++) {
	if (!strcmp(argv[i], "-c"))
	    do_class = 1;
	else {
	    struct stat buf;

	    stat(argv[i], &buf);
	    if (S_ISDIR(buf.st_mode)) {
		DIR *dir = opendir(argv[i]);
		struct dirent *de;

		while ((de = readdir(dir))) {
		    char fn[1024];

		    sprintf(fn, "%s/%s", argv[i], de->d_name);
		    stat(fn, &buf);
		    if (S_ISREG(buf.st_mode)) {
			pgmlist[n_pgm++] = load_pgm(strdup(fn));
		    }
		}
		closedir(dir);
	    } else {
		pgmlist[n_pgm++] = load_pgm(argv[i]);
	    }
	}
    }

    if (do_class) {
	for (i = 0; i < n_pgm; i++)
	    threshold(pgmlist[i]);
	return classify(pgmlist, n_pgm);
    }

    bb = new_blendbuf(pgmlist[0]->xs, pgmlist[0]->ys);
    for (i = 0; i < n_pgm; i++) {
	int xo, yo;
	int score = align_pgms(pgmlist[0], pgmlist[i], &xo, &yo);
	fprintf(stderr, "%s: score = %d, offset = %d, %d\n",
		pgmlist[i]->fn, score, xo, yo);
	add_pgm(bb, pgmlist[i], xo, yo);
    }

    print_blendbuf(bb, stdout);

    return 0;
}
